# -*- coding: utf-8 -*-

# pylint: disable=invalid-name

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""Conda Packages Table model."""

from qtpy.QtCore import QAbstractTableModel, QModelIndex, QSize, Qt  # pylint: disable=no-name-in-module
from qtpy.QtGui import QColor, QIcon
from anaconda_navigator.utils import constants as C
from anaconda_navigator.utils import sort_versions
from anaconda_navigator.utils.styles import MANAGER_TABLE_STYLES


GLOBAL_PALLETE = None


def process_pallete():
    """Turn the styles palette into QIcons or QColors for use in the model."""
    global GLOBAL_PALLETE  # pylint: disable=global-statement
    if GLOBAL_PALLETE is None:
        GLOBAL_PALLETE = {}
        for key, value in MANAGER_TABLE_STYLES.items():
            if key.startswith('icon.'):
                item = QIcon(value)
            elif key.startswith('color.'):
                item = QColor(value)
            elif key.startswith('size.'):
                item = QSize(*value)
            GLOBAL_PALLETE[key] = item
    return GLOBAL_PALLETE


class CondaPackagesModel(QAbstractTableModel):  # pylint: disable=too-many-public-methods
    """Abstract Model to handle the packages in a conda environment."""

    def __init__(self, parent, packages, data):
        """Abstract Model to handle the packages in a conda environment."""
        super().__init__(parent)
        self._parent = parent
        self._packages = packages
        self._rows = data
        self._name_to_index = {r[C.COL_NAME]: i for i, r in enumerate(data)}
        self._palette = process_pallete()

    def _update_cell(self, row, column):
        start = self.index(row, column)
        end = self.index(row, column)
        self.dataChanged.emit(start, end)

    @staticmethod
    def flags(index):
        """Override Qt method."""
        column = index.column()
        if index.isValid() and (column in [C.COL_START, C.COL_END]):
            return Qt.ItemFlags(Qt.ItemIsEnabled)
        return Qt.ItemFlags(Qt.ItemIsEnabled)

    def data(  # pylint: disable=too-many-branches,too-many-return-statements,too-many-statements
            self, index, role=Qt.DisplayRole,
    ):
        """Override Qt method."""
        if not index.isValid() or not 0 <= index.row() < len(self._rows):
            return None

        row = index.row()
        column = index.column()

        P = self._palette  # Include the palete directly here.

        if self._rows[row] == row:
            action = C.ACTION_NONE
            type_ = ''
            name = ''
            description = ''
            version = '-'
            status = -1
        else:
            action = self._rows[row][C.COL_ACTION]
            type_ = self._rows[row][C.COL_PACKAGE_TYPE]
            name = self._rows[row][C.COL_NAME]
            description = self._rows[row][C.COL_DESCRIPTION]
            version = self._rows[row][C.COL_VERSION]
            status = self._rows[row][C.COL_STATUS]

        is_upgradable = self.is_upgradable(self.index(row, C.COL_VERSION))

        if role == Qt.DisplayRole:
            if column == C.COL_PACKAGE_TYPE:
                return type_
            if column == C.COL_NAME:
                return name
            if column == C.COL_VERSION:
                return version
            if column == C.COL_STATUS:
                return status
            if column == C.COL_DESCRIPTION:
                return description
            if column == C.COL_ACTION:
                return action
        elif role == Qt.TextAlignmentRole:
            if column in [C.COL_NAME, C.COL_DESCRIPTION]:
                return int(Qt.AlignLeft | Qt.AlignVCenter)
            if column in [C.COL_VERSION]:
                return int(Qt.AlignLeft | Qt.AlignVCenter)
            return int(Qt.AlignCenter)

        elif role == Qt.DecorationRole:
            if column == C.COL_ACTION:
                if action == C.ACTION_NONE:
                    if status == C.NOT_INSTALLED:
                        return P['icon.action.not_installed']
                    if status in [C.UPGRADABLE, C.MIXGRADABLE]:
                        return P['icon.action.installed']
                    if status in [C.INSTALLED, C.DOWNGRADABLE, C.MIXGRADABLE]:
                        return P['icon.action.installed']
                elif action == C.ACTION_INSTALL:
                    return P['icon.action.add']
                elif action == C.ACTION_REMOVE:
                    return P['icon.action.remove']
                elif action == C.ACTION_UPGRADE:
                    return P['icon.action.upgrade']
                elif action == C.ACTION_UPDATE:
                    return P['icon.action.upgrade']
                elif action == C.ACTION_DOWNGRADE:
                    return P['icon.action.downgrade']
                else:
                    return None
            elif column == C.COL_PACKAGE_TYPE:
                if type_ == C.CONDA_PACKAGE:
                    return P['icon.anaconda']
                if type_ == C.PIP_PACKAGE:
                    return P['icon.python']
                return None
            elif column == C.COL_VERSION:
                if is_upgradable:
                    return P['icon.upgrade.arrow']
                return P['icon.spacer']

        elif role == Qt.ToolTipRole:
            if column == C.COL_PACKAGE_TYPE:
                if type_ == C.CONDA_PACKAGE:
                    return 'Conda package'
                if type_ == C.PIP_PACKAGE:
                    return 'Python package'
            elif (column == C.COL_VERSION) and is_upgradable:
                return 'Update available'

        elif role == Qt.ForegroundRole:
            if column in [C.COL_NAME, C.COL_DESCRIPTION]:
                if status in [C.INSTALLED, C.UPGRADABLE, C.DOWNGRADABLE, C.MIXGRADABLE]:
                    return QColor('#000')

                if status in [C.NOT_INSTALLED]:
                    return P['color.foreground.not.installed']

            elif (column in [C.COL_VERSION]) and is_upgradable:
                return P['color.foreground.upgrade']

        elif (role == Qt.SizeHintRole) and (column in [C.ACTION_COLUMNS, C.COL_PACKAGE_TYPE]):
            return P['size.icons']

        return None

    @staticmethod
    def headerData(section, orientation, role=Qt.DisplayRole):  # pylint: disable=too-many-return-statements
        """Override Qt method."""
        if orientation == Qt.Horizontal:
            if role == Qt.TextAlignmentRole:
                return int(Qt.AlignLeft | Qt.AlignVCenter)

            if (role == Qt.ToolTipRole) and (section == C.COL_PACKAGE_TYPE):
                return 'Package type: Conda, Pip'

            if role == Qt.DisplayRole:
                if section == C.COL_PACKAGE_TYPE:
                    return 'T'
                if section == C.COL_NAME:
                    return 'Name'
                if section == C.COL_VERSION:
                    return 'Version'
                if section == C.COL_DESCRIPTION:
                    return 'Description'
                if section == C.COL_STATUS:
                    return 'Status'
                return None

        return None

    def rowCount(self, index=QModelIndex()):  # pylint: disable=unused-argument
        """Override Qt method."""
        return len(self._rows)

    @staticmethod
    def columnCount(index=QModelIndex()):  # pylint: disable=unused-argument
        """Override Qt method."""
        return len(C.COLUMNS)

    def row(self, rownum):
        """Return the row data."""
        return self._rows[rownum]

    def first_index(self):
        """Return the first model index."""
        return self.index(0, 0)

    def last_index(self):
        """Return the last model index."""
        return self.index(self.rowCount() - 1, self.columnCount() - 1)

    def update_row_icon(self, row, column):
        """Update the model index icon."""
        if column in C.ACTION_COLUMNS:
            r = self._rows[row]
            actual_state = r[column]
            r[column] = not actual_state
            self._rows[row] = r
            self._update_cell(row, column)

    def is_installable(self, model_index):
        """Return if the pacakge index can be installed."""
        row = model_index.row()
        status = self._rows[row][C.COL_STATUS]
        return status == C.NOT_INSTALLED

    def is_removable(self, model_index):
        """Return if the installed pacakge index can be removed."""
        row = model_index.row()
        status = self._rows[row][C.COL_STATUS]
        return status in [C.UPGRADABLE, C.DOWNGRADABLE, C.INSTALLED, C.MIXGRADABLE]

    def is_upgradable(self, model_index):
        """Return if the installed pacakge index can be upgraded."""
        row = model_index.row()
        status = self._rows[row][C.COL_STATUS]
        return status in [C.UPGRADABLE, C.MIXGRADABLE]

    def is_downgradable(self, model_index):
        """Return if the installed pacakge index can be downgraded."""
        row = model_index.row()
        status = self._rows[row][C.COL_STATUS]
        return status in [C.DOWNGRADABLE, C.MIXGRADABLE]

    def action_status(self, model_index):
        """Return the current action status."""
        row = model_index.row()
        action_status = self._rows[row][C.COL_ACTION]
        return action_status

    def set_action_status(self, model_index, status, version=None):
        """Set index status action."""
        row = model_index.row()
        self._rows[row][C.COL_ACTION] = status
        self._rows[row][C.COL_ACTION_VERSION] = version
        self._update_cell(row, model_index.column())

    def clear_actions(self):
        """Clear the selected conda actions."""
        for i, row in enumerate(self._rows):
            row[C.COL_ACTION] = C.ACTION_NONE
            row[C.COL_ACTION_VERSION] = None
            self._update_cell(i, C.COL_ACTION)
            self._update_cell(i, C.COL_ACTION_VERSION)

    def count_remove_actions(self):
        """Return the number of conda remove actions selected."""
        count = 0
        for row in self._rows:
            action = row[C.COL_ACTION]
            type_ = row[C.COL_PACKAGE_TYPE]
            if (action == C.ACTION_REMOVE) and (type_ == C.CONDA_PACKAGE):
                count += 1
        return count

    def count_install_actions(self):
        """Return number of install/upgrade/downgrade actions selected."""
        count = 0
        for row in self._rows:
            type_ = row[C.COL_PACKAGE_TYPE]
            action = row[C.COL_ACTION]
            if (action in [C.ACTION_DOWNGRADE, C.ACTION_INSTALL, C.ACTION_UPGRADE]) and (type_ == C.CONDA_PACKAGE):
                count += 1
        return count

    def count_update_actions(self):
        """Return number of update (no version given) actions selected."""
        count = 0
        for row in self._rows:
            type_ = row[C.COL_PACKAGE_TYPE]
            action = row[C.COL_ACTION]
            if (action in [C.ACTION_UPDATE]) and (type_ == C.CONDA_PACKAGE):
                count += 1
        return count

    def get_actions(self):
        """Return the selected conda actions."""
        dic = {
            C.CONDA_PACKAGE: {
                C.ACTION_INSTALL: [],
                C.ACTION_REMOVE: [],
                C.ACTION_UPGRADE: [],
                C.ACTION_DOWNGRADE: [],
                C.ACTION_UPDATE: [],
            },
            C.PIP_PACKAGE: {
                C.ACTION_REMOVE: [],
            }
        }

        for row in self._rows:
            action = row[C.COL_ACTION]
            name = row[C.COL_NAME]
            type_ = row[C.COL_PACKAGE_TYPE]
            action_version = row[C.COL_ACTION_VERSION]
            current_version = self.get_package_version(name)

            if action != C.ACTION_NONE:
                version_from = current_version
                version_to = action_version
                dic[type_][action].append({
                    'name': name,
                    'version_from': version_from,
                    'version_to': version_to,
                })
        return dic

    def get_action_count(self):
        """Count selected actions."""
        count = 0
        for row in self._rows:  # pylint: disable=consider-using-enumerate
            action = row[C.COL_ACTION]

            if action != C.ACTION_NONE:
                count += 1
        return count

    def get_package_versions(self, name):
        """
        Return the package canonical name for a given package `name`.

        name : str
            Name of the package
        """
        package_data = self._packages.get(name)
        versions = []

        if package_data:
            versions = sort_versions(list(package_data.get('versions', [])))

        return versions

    def get_package_version(self, name):
        """Return the package version for a given package `name`."""
        if name in self._name_to_index:
            index = self._name_to_index[name]
            version = self.row(index)[C.COL_VERSION]
            return version.replace(C.UPGRADE_SYMBOL, '')
        return ''



# -*- coding: utf8 -*-

"""Utilities for version comparison."""

from __future__ import annotations

__all__ = ['compare']

import re
import typing

import packaging.version


NUMBER: typing.Final[typing.Pattern[str]] = re.compile(r'([0-9]+)')


class SupportsLessThan(typing.Protocol):  # pylint: disable=too-few-public-methods
    """Value that can be compared to another one."""

    def __lt__(self, other: typing.Any) -> bool:
        """Check if value is less than other one."""


SupportsLessThanT = typing.TypeVar('SupportsLessThanT', bound=SupportsLessThan)


def _compare(first: SupportsLessThanT, second: SupportsLessThanT) -> int:
    """
    Compare two values.

    :return: :code:`1` if :code:`first` > :code:`second`; :code:`-1` if :code:`first` < :code:`second`; :code:`0` if
             both values are equal.
    """
    if first < second:
        return -1
    if second < first:
        return 1
    return 0


def _parse(value: str) -> typing.Tuple[typing.Union[int, str], ...]:
    """Parse any version-like string into tuple of integer-string values."""
    return tuple(
        int(part) if (index % 2) else str(part)
        for index, part in enumerate(NUMBER.split(value))
    )


def compare(first: str, second: str) -> int:
    """Compare two strings with versions."""
    try:
        return _compare(packaging.version.Version(first), packaging.version.Version(second))
    except packaging.version.InvalidVersion:
        return _compare(_parse(first), _parse(second))



# -*- coding: utf-8 -*-

"""Common configuration structures to use in :mod:`~anaconda_navigator.config.preferences`."""

from __future__ import annotations

__all__ = ['SidebarLink', 'SidebarSocial', 'Interval', 'Intervals']

import bisect
import typing

import attr

T = typing.TypeVar('T')
T_co = typing.TypeVar('T_co', covariant=True)


class SidebarLink(typing.NamedTuple):
    """Description of a single link in a sidebar."""

    text: str
    url: str
    utm_medium: str


class SidebarSocial(typing.NamedTuple):
    """Description of a single social button in a sidebar."""

    text: str
    url: str


@attr.s(auto_attribs=True, eq=False, frozen=True, slots=True)
class Interval(typing.Generic[T_co]):  # pylint: disable=too-few-public-methods
    """
    Value that repeats `count` times.

    :param count: how many times the value should be repeated.
    :param value: value that should be repeated.
    """

    count: int
    value: T_co


class Intervals(typing.Generic[T]):
    """
    Sequence of repeating values.

    :param args: values to store in this sequence.
    :param offset: index of the first value in the sequence. This allows creating custom sequences, which may start
                   from any custom index (e.g. create sequence with indices from -5 to 15, or from 100 to 123).
    """

    __slots__ = ('__keys', '__offset', '__values')

    def __init__(self, *args: Interval[T], offset: int = 0) -> None:
        """Initialize new :class:`~Intervals` instance."""
        if not args:
            raise TypeError(f'at least single {Interval.__name__} must be provided')

        self.__keys: typing.Final[typing.List[int]] = []
        self.__offset: typing.Final[int] = offset
        self.__values: typing.Final[typing.List[T]] = []

        offset -= 1

        arg: Interval[T]
        for arg in args:
            if arg.count <= 0:
                raise ValueError(f'each item may repeat only a positive number of times, not {arg.count}')

            offset += arg.count
            self.__keys.append(offset)
            self.__values.append(arg.value)

    @property
    def first(self) -> int:  # noqa: D401
        """Index of the first element in the sequence."""
        return self.__offset

    @property
    def last(self) -> int:  # noqa: D401
        """Index of the last element in the sequence."""
        return self.__keys[-1]

    def get(self, index: int, *, strict: bool = False) -> T:
        """
        Retrieve element at `index`.

        If `strict` is set - method may raise an :exc:`~IndexError` if invalid item is requested. Otherwise - method
        will return value closest to the requested index.
        """
        if strict:
            if index < self.__offset:
                raise IndexError(f'index must be at least {self.__offset!r}')
            if index > self.__keys[-1]:
                raise IndexError(f'index must be at most {self.__offset!r}')

        index = bisect.bisect_left(self.__keys, index, 0, len(self.__keys) - 1)
        return self.__values[index]



# -*- coding: utf-8 -*-

# pylint: disable=invalid-name

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""
Transitional module providing compatibility functions between python 2 and 3.

This module should be fully compatible with:
    * Python >=v2.7
    * Python 3
"""

from __future__ import print_function

import typing
import operator
import os
import sys


PY2 = sys.version[0] == '2'
PY3 = sys.version[0] == '3'

# =============================================================================
# Data types
# =============================================================================
TEXT_TYPES: typing.Tuple[type, ...] = (str,)
INT_TYPES: typing.Tuple[type, ...] = (int,)
NUMERIC_TYPES: typing.Tuple[type, ...] = INT_TYPES + (float, complex)


# =============================================================================
# Strings
# =============================================================================

def is_text_string(obj):
    """
    Return True if `obj` is a text string, False if it is anything else.

    Binary data (Python 3) or QString (Python 2, PyQt API #1).
    """
    return isinstance(obj, str)


def is_binary_string(obj):
    """Return True if obj is a binary string, False if it is anything else."""
    return isinstance(obj, bytes)


def is_string(obj):
    """
    Check if object is a string.

    Return True if `obj` is a text or binary Python string object False if it
    is anything else, like a QString (Python 2, PyQt API #1).
    """
    return is_text_string(obj) or is_binary_string(obj)


def is_unicode(obj):
    """Return True if `obj` is unicode."""
    return isinstance(obj, str)


def to_text_string(obj, encoding=None):
    """Convert `obj` to (unicode) text string."""
    if encoding is None:
        return str(obj)
    if isinstance(obj, str):  # In case this function is not used properly, this could happen
        return obj
    return str(obj, encoding)


def to_binary_string(obj, encoding=None):
    """Convert `obj` to binary string (bytes in Python 3, str in Python 2)."""
    return bytes(obj, 'utf-8' if encoding is None else encoding)


def u(obj):
    """Return string as it is."""
    return obj


# =============================================================================
# Misc.
# =============================================================================

getcwd = os.getcwd
str_lower = str.lower


def cmp(a, b):
    """Return negative if a<b, zero if a==b, positive if a>b."""
    return (a > b) - (a < b)


def qbytearray_to_str(qba):
    """Convert QByteArray object to str in a way compatible with Python 2/3."""
    return str(bytes(qba.toHex().data()).decode())


# =============================================================================
# Dict funcs
# =============================================================================

def iterkeys(d, **kw):
    """Return an iterator over the dictionary's keys."""
    return iter(d.keys(**kw))


def itervalues(d, **kw):
    """Return an iterator over the dictionary's values."""
    return iter(d.values(**kw))


def iteritems(d, **kw):
    """Return an iterator over the dictionary's items."""
    return iter(d.items(**kw))


def iterlists(d, **kw):
    """Return an iterator over a multi dictionary."""
    return iter(d.lists(**kw))


viewkeys = operator.methodcaller('keys')

viewvalues = operator.methodcaller('values')

viewitems = operator.methodcaller('items')



# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""
Home Tab.

This widget does not perform the actual actions but it emits signals that
should be connected to the final controller on the main window.
"""

from __future__ import absolute_import, division, print_function

__all__ = ['HomeTab']

import typing

from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QApplication, QHBoxLayout, QLabel, QProgressBar, QVBoxLayout

from anaconda_navigator.api.anaconda_api import AnacondaAPI
from anaconda_navigator.api import types as api_types
from anaconda_navigator.utils import constants
from anaconda_navigator.utils.analytics import GATracker
from anaconda_navigator.widgets import common as global_commons
from anaconda_navigator.widgets import (
    ButtonNormal, ComboBoxBase, FrameTabContent, FrameTabFooter, FrameTabHeader, LabelBase, SpacerHorizontal, WidgetBase
)
from anaconda_navigator.widgets.lists.apps import ListItemApplication, ListWidgetApplication


ApplicationFilter = typing.Literal['all', 'installed', 'uninstalled', 'updatable']


# --- Application filters
# -----------------------------------------------------------------------------

class ApplicationFilterDetails(typing.NamedTuple):
    """Additional details for the application filters."""

    verbose_name: str

    filter_function: typing.Callable[['api_types.Application'], bool]


class CheckInstalled:  # pylint: disable=too-few-public-methods
    """
    Common checker for application installation state.

    :param target: Expected installation state of an application.
    """

    __slots__ = ('__target',)

    def __init__(self, target: bool) -> None:
        """Initialize new :class:`~CheckInstalled` instance."""
        self.__target: typing.Final[bool] = target

    @property
    def target(self) -> bool:  # noqa: D401
        """Expected installation state."""
        return self.__target

    def __call__(self, application: 'api_types.Application') -> bool:
        """Check if application installation state."""
        return application.get('installed', False) is self.__target


def check_updatable(application: 'api_types.Application') -> bool:
    """Check if there is an update for an `application`."""
    version: typing.Optional[str] = application.get('version', None)
    versions: typing.Sequence[str] = application.get('versions', ())
    return application.get('installed', False) and bool(version) and bool(versions) and (version != versions[-1])


APPLICATION_FILTERS: typing.Final[typing.Mapping[ApplicationFilter, ApplicationFilterDetails]] = {
    'all': ApplicationFilterDetails('All applications', lambda application: True),
    'installed': ApplicationFilterDetails('Installed applications', CheckInstalled(True)),
    'uninstalled': ApplicationFilterDetails('Not installed applications', CheckInstalled(False)),
    'updatable': ApplicationFilterDetails('Updatable applications', check_updatable),
}


def application_sorting_key(application: 'api_types.Application') -> typing.Tuple[int, bool, int, str]:
    """Prepare sorting key for an application."""
    rank: int = -application.get('rank', 0)
    installed: bool = not application.get('installed', False)

    category: int
    application_type: typing.Optional[constants.AppType] = application.get('app_type', None)
    if application_type in (constants.AppType.INSTALLABLE, constants.AppType.CONDA):
        category = 0
    elif application_type == constants.AppType.WEB:
        category = 1
    else:
        category = 2

    display_name: str = application.get('display_name', '').lower()

    return rank, installed, category, display_name


# --- Custom widgets used with CSS styling
# -----------------------------------------------------------------------------

class ButtonHomeRefresh(global_commons.IconButton):  # pylint: disable=too-few-public-methods
    """QFrame used for CSS styling refresh button inside the Home Tab."""


class ComboHomeFilter(ComboBoxBase):  # pylint: disable=too-few-public-methods
    """Widget Used for CSS styling."""


class ComboHomeEnvironment(ComboBoxBase):  # pylint: disable=too-few-public-methods
    """Widget Used for CSS styling."""


class ButtonHomeChannels(ButtonNormal):
    """Widget Used for CSS styling."""


class LabelHome(LabelBase):
    """QLabel used for CSS styling the Home Tab label."""


# --- Main widget
# -----------------------------------------------------------------------------
class HomeTab(WidgetBase):  # pylint: disable=too-many-instance-attributes
    """Home applications tab."""
    # name, prefix, sender
    sig_item_selected = Signal(object, object, object)

    # button_widget, sender
    sig_channels_requested = Signal(object, object)

    # application_name, command, extra_arguments, prefix, leave_path_alone, sender, non_conda, app_type
    sig_launch_action_requested = Signal(object, object, object, bool, object, object, object, object)

    # action, application_name, version, sender, non_conda, app_type
    sig_conda_action_requested = Signal(object, object, object, object, object, object)

    # url
    sig_url_clicked = Signal(object)

    # NOTE: Connect these signals to have more granularity
    # [{'name': package_name, 'version': version}...], sender
    sig_install_action_requested = Signal(object, object)
    sig_remove_action_requested = Signal(object, object)

    def __init__(self, parent=None):  # pylint: disable=too-many-statements
        """Home applications tab."""
        super().__init__(parent)

        # Variables
        self._parent = parent
        self.api = AnacondaAPI()
        self.applications = None
        self.app_timers = None
        self.current_prefix = None
        self.tracker = GATracker()

        self.__applications: typing.Mapping['api_types.ApplicationName', 'api_types.Application'] = {}
        self.__applications_filter: 'ApplicationFilter' = 'all'

        # Widgets
        self.list = ListWidgetApplication()
        self.button_channels = ButtonHomeChannels('Channels')
        self.button_refresh = ButtonHomeRefresh()
        self.combo_filter = ComboHomeFilter()
        self.label_home = LabelHome('on')
        self.combo_environment = ComboHomeEnvironment()
        self.frame_top = FrameTabHeader(self)
        self.frame_body = FrameTabContent(self)
        self.frame_bottom = FrameTabFooter(self)
        self.label_status_action = QLabel('')
        self.label_status = QLabel('')
        self.progress_bar = QProgressBar()
        self.first_widget = self.combo_environment
        self.te_alert = global_commons.TeamEditionServerAlert()

        # Widget setup
        self.setObjectName('Tab')
        self.progress_bar.setTextVisible(False)
        self.list.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)

        application_filter: 'ApplicationFilter'
        application_filter_details: ApplicationFilterDetails
        for application_filter, application_filter_details in APPLICATION_FILTERS.items():
            self.combo_filter.addItem(application_filter_details.verbose_name, application_filter)
        self.combo_filter.setCurrentText(APPLICATION_FILTERS[self.__applications_filter].verbose_name)

        # Layout
        layout_top = QHBoxLayout()
        layout_top.addWidget(self.combo_filter)
        layout_top.addWidget(SpacerHorizontal())
        layout_top.addWidget(self.label_home)
        layout_top.addWidget(SpacerHorizontal())
        layout_top.addWidget(self.combo_environment)
        layout_top.addWidget(SpacerHorizontal())
        layout_top.addWidget(self.button_channels)
        layout_top.addWidget(SpacerHorizontal())
        layout_top.addStretch()
        layout_top.addWidget(self.button_refresh)
        self.frame_top.setLayout(layout_top)

        layout_body = QVBoxLayout()
        layout_body.addWidget(self.te_alert)
        layout_body.addWidget(self.list)
        self.frame_body.setLayout(layout_body)

        layout_bottom = QHBoxLayout()
        layout_bottom.addWidget(self.label_status_action)
        layout_bottom.addWidget(SpacerHorizontal())
        layout_bottom.addWidget(self.label_status)
        layout_bottom.addStretch()
        layout_bottom.addWidget(self.progress_bar)
        self.frame_bottom.setLayout(layout_bottom)

        layout = QVBoxLayout()
        layout.addWidget(self.frame_top)
        layout.addWidget(self.frame_body)
        layout.addWidget(self.frame_bottom)
        self.setLayout(layout)

        # Signals
        self.combo_filter.currentIndexChanged.connect(self._filter_selected)
        self.list.sig_conda_action_requested.connect(self.sig_conda_action_requested)
        self.list.sig_url_clicked.connect(self.sig_url_clicked)
        self.list.sig_launch_action_requested.connect(self.sig_launch_action_requested)
        self.button_channels.clicked.connect(self.show_channels)
        self.button_refresh.clicked.connect(self.refresh_cards)
        self.progress_bar.setVisible(False)

    # --- Setup methods
    # -------------------------------------------------------------------------
    def setup(self, conda_data):
        """Setup the tab content."""
        conda_processed_info = conda_data.get('processed_info')
        environments = conda_processed_info.get('__environments')
        self.current_prefix = conda_processed_info.get('default_prefix')
        self.set_environments(environments)

        self.__applications = conda_data.get('applications')
        self.update_applications()

    def set_environments(self, environments):
        """Setup the environments list."""
        # Disconnect to avoid triggering the signal when updating the content
        try:
            self.combo_environment.currentIndexChanged.disconnect()
        except TypeError:
            pass

        self.combo_environment.clear()

        font_metrics = self.combo_environment.fontMetrics()
        widths = []

        index: int
        for index, (env_prefix, env_name) in enumerate(environments.items()):
            widths.append(font_metrics.width(env_name))
            self.combo_environment.addItem(env_name, env_prefix)
            self.combo_environment.setItemData(index, env_prefix, Qt.ToolTipRole)

        for index, (env_prefix, env_name) in enumerate(environments.items()):
            if self.current_prefix == env_prefix:
                break
        else:
            index = 0

        self.combo_environment.setCurrentIndex(index)
        self.combo_environment.currentIndexChanged.connect(self._environment_selected)

        # Fix combobox width
        width = max(widths) + 64
        self.combo_environment.setMinimumWidth(width)

    def update_applications(self) -> None:
        """Build the list of applications present in the current conda env."""
        processed_applications: typing.List['api_types.Application'] = sorted(
            filter(
                APPLICATION_FILTERS[self.__applications_filter].filter_function,
                self.api.process_apps(self.__applications, prefix=self.current_prefix).values(),
            ),
            key=application_sorting_key,
        )

        self.list.clear()

        application_data: 'api_types.Application'
        for application_data in processed_applications:
            self.list.addItem(ListItemApplication(prefix=self.current_prefix, **application_data))

        self.list.update_style_sheet()
        self.set_widgets_enabled(True)
        self.update_status()

    # --- Other methods
    # -------------------------------------------------------------------------
    def current_environment(self):
        """Return the current selected environment."""
        env_name: str = self.combo_environment.currentText()
        return self.api.conda_get_prefix_envname(env_name)

    def refresh_cards(self):
        """Refresh application widgets.

        List widget items sometimes are hidden on resize. This method tries
        to compensate for that refreshing and repainting on user demand.
        """
        self.tracker.track_event('home', 'click', 'Refresh')
        self.list.update_style_sheet()
        self.list.repaint()
        for item in self.list.items():
            if not item.widget.isVisible():
                item.widget.repaint()

        worker = self.api.conda_data(prefix=self.current_prefix)
        worker.sig_chain_finished.connect(lambda _, output, __: self.setup(output))
        self.update_status(action='Refreshing applications', value=0, max_value=0)

    def show_channels(self):
        """Emit signal requesting the channels dialog editor."""
        self.tracker.track_event('home', 'click', 'Channels')
        self.sig_channels_requested.emit(self.button_channels, constants.TAB_HOME)

    # --- Common Helpers (# NOTE: factor out to common base widget)
    # -------------------------------------------------------------------------
    def _environment_selected(self, index: int) -> None:
        """Notify that the item in combo (environment) changed."""
        name: str = self.combo_environment.itemText(index)
        prefix: str = self.combo_environment.itemData(index)
        self.sig_item_selected.emit(name, prefix, constants.TAB_HOME)

    def _filter_selected(self, index: int) -> None:
        """Process change of application filter."""
        self.applications_filter = self.combo_filter.itemData(index)

    @property
    def last_widget(self):
        """Return the last element of the list to be used in tab ordering."""
        if self.list.items():
            return self.list.items()[-1].widget
        return None

    @property
    def applications_filter(self) -> 'ApplicationFilter':  # noqa: D401
        """Current filter applied to the application list."""
        return self.__applications_filter

    @applications_filter.setter
    def applications_filter(self, value: 'ApplicationFilter') -> None:
        """Update `application_filter` value."""
        if self.__applications_filter == value:
            return

        self.combo_filter.setCurrentText(APPLICATION_FILTERS[value].verbose_name)
        self.__applications_filter = value
        self.update_applications()

    def ordered_widgets(self, next_widget=None):  # pylint: disable=unused-argument
        """Return a list of the ordered widgets."""
        ordered_widgets = [
            self.combo_filter,
            self.combo_environment,
            self.button_channels,
            self.button_refresh,
        ]
        ordered_widgets += self.list.ordered_widgets()

        return ordered_widgets

    def set_widgets_enabled(self, value: bool) -> None:
        """Enable or disable widgets."""
        self.combo_filter.setEnabled(value)
        self.combo_environment.setEnabled(value)
        self.button_channels.setEnabled(value)
        self.button_refresh.setEnabled(value)

        for item in self.list.items():
            item.button_install.setEnabled(value)
            item.button_options.setEnabled(value)

            if value:
                item.set_loading(not value)

    def update_items(self):
        """Update status of items in list."""
        if self.list:
            for item in self.list.items():
                item.update_status()

    def update_status(self, action='', message='', value=None, max_value=None):
        """Update the application action status."""

        # Elide if too big
        width = QApplication.desktop().availableGeometry().width()
        max_status_length = round(width * (2.0 / 3.0), 0)
        msg_percent = 0.70

        font_metrics = self.label_status_action.fontMetrics()
        action = font_metrics.elidedText(action, Qt.ElideRight, round(max_status_length * msg_percent))
        message = font_metrics.elidedText(message, Qt.ElideRight, round(max_status_length * (1 - msg_percent)))
        self.label_status_action.setText(action)
        self.label_status.setText(message)

        if max_value is None and value is None:
            self.progress_bar.setVisible(False)
        else:
            self.progress_bar.setVisible(True)
            self.progress_bar.setMaximum(max_value)
            self.progress_bar.setValue(value)

    def update_style_sheet(self):
        """Update custom CSS style sheet."""
        self.list.update_style_sheet()



# -*- coding: utf-8 -*-

"""Extra solvers to migrate trusted_servers to domain-only format."""

from __future__ import annotations

__all__ = ()

import configparser
import contextlib
import html
import typing
from . import core


class MigrationItem(typing.NamedTuple):
    """Instructions on how a preference should be migrated."""

    section: str
    from_option: str
    to_option: str


MIGRATION: typing.Final[typing.Sequence[MigrationItem]] = (
    MigrationItem(section='main', from_option='team_edition_api_url', to_option='anaconda_server_api_url'),
    MigrationItem(section='main', from_option='team_edition_token', to_option='anaconda_server_token'),
    MigrationItem(section='main', from_option='team_edition_token_id', to_option='anaconda_server_token_id'),
    MigrationItem(section='main', from_option='commercial_edition_url', to_option='anaconda_professional_url'),
)


@core.CONFIGURATION_POOL.register
def migrate_branding(context: core.ConfigurationContext) -> typing.Optional[core.SolvedError]:
    """Migrate options affected by branding changes."""
    item: MigrationItem
    issues: typing.List[MigrationItem] = []
    for item in MIGRATION:
        new_value: typing.Optional[str]
        try:
            new_value = context.config.get(item.section, item.from_option)
        except configparser.NoOptionError:
            pass
        else:
            with contextlib.suppress(configparser.NoOptionError):
                default: typing.Any = context.config.get_default(item.section, item.to_option)
                if context.config.get(item.section, item.to_option) != default:
                    issues.append(item)
            context.config.set(item.section, item.to_option, new_value)
            context.config.remove_option(item.section, item.from_option)

    if issues:
        return core.SolvedError(
            caption='Conflicting preferences in navigator configuration',
            message='<br>'.join(
                f'<b>{html.escape(issue.to_option)}</b> was overwritten by <b>{html.escape(issue.from_option)}</b>'
                for issue in issues
            ),
            tags='branding_overlap',
        )
    return None
